/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveConsole.cpp
 * Creation date : 15/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveConsole.h"
#include "tveTexture.h"
#include <stdarg.h>

#define NUM_CHAR_PER_LINE ((int)(2*width/CON_CHAR_WIDTH)-2)

#define DEBUG_TO_FILE
#define DEBUG

//=====================================================================
// tveConsole
//=====================================================================

/*
 *
 */
tveConsole::tveConsole(int speed)
{
#ifdef DEBUG_TO_FILE
    char filename[256];
    sprintf(filename, "%s%s", tveVoyager::path, "debug.txt");
    f = fopen(filename, "wt");
    fprintf(f, "Debug file for Terrain Voyager Engine\n");
#endif
    this->speed   = -speed;
    ypos          = 0;
    last_buf_pos  = 0;
    last_his_pos  = 0;
    line_pos      = 0;
    scr_pos       = 0;
    line[0]       = '\0';
    history[0]    = strdup(line);
    his_pos       = 0;
    buffer[0]     = line;
    cmd_idx       = 0;
    var_idx       = 0;
    tex_font      = 0;
    scr_buf_idx   = 0;
    command_queue = new tveQueue();
    
    // predefined commands
    addCommand(new tveConsoleCommand("echo", GC_ECHO, ::genericCommand));
    addCommand(new tveConsoleCommand("cls",  GC_CLS,  ::genericCommand));
    addCommand(new tveConsoleCommand("exec", GC_EXEC, ::genericCommand));
    addCommand(new tveConsoleCommand("help", GC_HELP, ::genericCommand));
    addCommand(new tveConsoleCommand("set",  GC_SET,  ::genericCommand));
    addCommand(new tveConsoleCommand("wait", GC_WAIT|CT_DELAYED, ::genericCommand));
}


/*
 *
 */
tveConsole::~tveConsole()
{
#ifdef DEBUG_TO_FILE
    fclose(f);
#endif
    delete command_queue;
    //TODO: unalloc buffer and history
}


/*
 *
 */
void tveConsole::start()
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, width, 0, height);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_DEPTH_TEST);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex_font);
    glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE);
}


/*
 *
 */
void tveConsole::stop()
{
    //FIXME: restore old blend function like it's done in Doom Legacy with XOR
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPopAttrib();
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
}


/*
 *
 */
void tveConsole::printf(int x, int y, char *lpFmt, ...)
{
    static char str[8192];
    va_list arglist;

    va_start(arglist, lpFmt);
    vsprintf(str, lpFmt, arglist);
    va_end(arglist);

    for (int i=0; str[i]!='\0'; i++)
    {
        if (!(i%NUM_CHAR_PER_LINE))
        {
            glLoadIdentity();
            glTranslatef((float)x, (float)y-CON_CHAR_HEIGHT*i/NUM_CHAR_PER_LINE, 0.0f);
            glScalef(0.5f, 0.5f, 1.0f);
        }
        glCallList(lists[str[i]-32]);
        glTranslatef(CON_CHAR_WIDTH, 0.0f, 0.0f);
    }
}


/*
 *
 */
void tveConsole::printf(int x, int y, Uint32 col, GLfloat scale, char *lpFmt, ...)
{
    static char str[8192];
    va_list arglist;

    va_start(arglist, lpFmt);
    vsprintf(str, lpFmt, arglist);
    va_end(arglist);
    screen_buffer[scr_buf_idx].x = x;
    screen_buffer[scr_buf_idx].y = y;
    screen_buffer[scr_buf_idx].col = col;
    screen_buffer[scr_buf_idx].scale = scale;
    screen_buffer[scr_buf_idx].text = strdup(str);
    scr_buf_idx++;
}


/*
 *
 */
void tveConsole::printBuffer()
{
    for (int j=0; j<scr_buf_idx; j++)
    {
        float x = (float)screen_buffer[j].x;
        float y = (float)screen_buffer[j].y;
        Uint32 col = screen_buffer[j].col;
        float scale = screen_buffer[j].scale;
        char  *str = screen_buffer[j].text;

        glColor4ub((col>>24)&0xff, (col>>16)&0xff, (col>>8)&0xff, col&0xff );
        glLoadIdentity();
        glTranslatef(x, y, 0.0f);
        glScalef(scale, scale, 1.0f);
        for (int i=0; str[i]!='\0'; i++)
        {
            glCallList(lists[str[i]-32]);
            glTranslatef(CON_CHAR_WIDTH, 0.0f, 0.0f);
        }
        free(str);
    }
    scr_buf_idx = 0;
}


/*
 *
 */
void tveConsole::flushBuffer()
{
    for (int j=0; j<scr_buf_idx; j++)
    {
        free(screen_buffer[j].text);
    }
    scr_buf_idx = 0;
}


/*
 *
 */
void tveConsole::printf(char *lpFmt, ...)
{
    static char str[CON_MAX_LINE];
    va_list arglist;

    va_start(arglist, lpFmt);
    vsprintf(str, lpFmt, arglist);
    va_end(arglist);

    buffer[last_buf_pos++] = strdup(str); //TODO: manage properly the circular buffer
    buffer[last_buf_pos] = line;
    scr_pos = last_buf_pos;
#ifdef DEBUG_TO_FILE
    fprintf(f, "%s\n", str);
#endif
#ifdef DEBUG
    ::printf("%s\n", str);
#endif
}


/*
 *
 */
void tveConsole::display()
{
    float ratio = tveVoyager::ratio*0.05f;
    // update and clamp ypos in the interval of [0,max_height]
    ypos = MAX(ypos+speed*ratio, 0.0f);
    ypos = MIN(ypos, max_height);
    if (visible())
    {
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA); //FIXME: change blend func when we really need it
        glColor4f(0.0f, 0.0f, 0.0f, 0.5f);
        glDisable(GL_TEXTURE_2D);
        glLoadIdentity();
        glBegin(GL_TRIANGLE_STRIP);
          glVertex2f(0.0f,         (float)height);
          glVertex2f(0.0f,         height-ypos);
          glVertex2f((float)width, (float)height);
          glVertex2f((float)width, height-ypos);
        glEnd();
        glEnable(GL_TEXTURE_2D);
        glColor4f(0.0f, 1.0f, 1.0f, 1.0f);
        glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE);
        // print ypos/CON_CHAR_HEIGHT lines on the screen with management of multi line
        for (int i=0, scr_pos=0; (scr_pos<ypos/CON_CHAR_HEIGHT) && (this->scr_pos >= i); i++)
        {
            scr_pos += 1+strlen(buffer[this->scr_pos-i])/NUM_CHAR_PER_LINE;
            printf(10, (height-(int)ypos)+(scr_pos-1)*CON_CHAR_HEIGHT, "%s", buffer[this->scr_pos-i]);
        }
        // show caret
        if (this->scr_pos==last_buf_pos)
        {
            printf(10+10*(strlen(buffer[last_buf_pos])%NUM_CHAR_PER_LINE), (height-(int)ypos), "_");
        }
    }
}


/*
 *
 */
bool tveConsole::visible()
{
    return (ypos > 0.0f);
}


/*
 *
 */
void tveConsole::handleCommands()
{
    tveConsoleCommandItem *command_item = (tveConsoleCommandItem *)command_queue->getFirst();

    if (command_item)
    {
        if (command_item->execute(this, command_item->tag, command_item->args))
        {
            command_queue->deQueue();
        }
    }
}


/*
 *
 */
void tveConsole::handleKeyDown(SDL_keysym keysym)
{
    if (speed < 0)
    {
        if ( (keysym.sym==SDLK_BACKQUOTE) || (keysym.sym==tveVoyager::cv_consolekey->value.integer) )
        {
            speed = ABS(speed);
            SDL_EnableUNICODE(1);
            SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY/2, SDL_DEFAULT_REPEAT_INTERVAL/2);
        }
        return;
    }
    switch (keysym.sym)
    {
        case SDLK_PAGEUP:
            //TODO: manage circular buffer
            if (scr_pos > 0)
            {
                scr_pos--;
            }
            break;
        case SDLK_PAGEDOWN:
            //TODO: manage circular buffer
            if (scr_pos < last_buf_pos)
            {
                scr_pos++;
            }
            break;
        case SDLK_DOWN:
            if (his_pos < last_his_pos-1)
            {
                his_pos++;
                strcpy(line, history[his_pos]);
                line_pos = strlen(line);
            }
            else
            {
                line[0] = '\0';
                line_pos = 0;
            }
            break;
        case SDLK_UP:
            if (his_pos > 0)
            {
                his_pos--;
            }
            strcpy(line, history[his_pos]);
            line_pos = strlen(line);
            break;
        case SDLK_BACKSPACE:
            line_pos = MAX(line_pos-1, 0);
            line[line_pos] = '\0';
            break;
        case SDLK_TAB:
            // do variable/command completion
            break;
        case SDLK_RETURN:
            {
                printf("%s", line);
                addHistory();
                parseCommand(line, false);
                line_pos = 0;
                line[line_pos] = '\0';
                buffer[last_buf_pos] = line;
            }
            break;
        case SDLK_BACKQUOTE:
            speed = -ABS(speed);
            SDL_EnableUNICODE(0);
            SDL_EnableKeyRepeat(0, 0);
            break;
        default:
            if (keysym.sym == tveVoyager::cv_consolekey->value.integer)
            {
                speed = -ABS(speed);
                SDL_EnableUNICODE(0);
                SDL_EnableKeyRepeat(0, 0);
                break;
            }
            char ch = keysym.unicode & 0x7F;
            if (ch >= 32) // printable
            {
                line[line_pos++] = ch;
                line[line_pos] = '\0';
            }
            break;
    }
}


/*
 *
 */
void tveConsole::addHistory()
{
    history[last_his_pos] = strdup(line);
    last_his_pos = (last_his_pos+1)%CON_HIS_SIZE;
    his_pos = last_his_pos;
}


/*
 *
 */
void tveConsole::addCommand(tveConsoleCommand *command)
{
    commands[cmd_idx++] = command;
}


/*
 *
 */
void tveConsole::addVariable(tveConsoleVariable *variable)
{
    variables[var_idx++] = variable;
}


/*
 *
 */
char *tveConsole::getCmd(char **s)
{
    while (**s == ' ')
    {
        (*s)++;
    }
    char *res = *s;
    while ( (**s != '\0') && (**s != ' ') )
    {
        (*s)++;
    }
    if (**s == ' ')
    {
        **s = '\0';
        (*s)++;
    }
    return res;
}


/*
 *
 */
void tveConsole::parseCommand(char *s, bool conf)
{
    char *cmd = getCmd(&s);
    int  len = strlen(cmd);

    if (!len)
    {
        return;
    }
    for (int i=0; i<cmd_idx; i++)
    {
        if (!strncasecmp(commands[i]->name, cmd, len))
        {
            if (commands[i]->tag & CT_DELAYED)
            {
                command_queue->enQueue(new tveConsoleCommandItem(commands[i]->name, commands[i]->tag, commands[i]->execute, s));
            }
            else
            {
                commands[i]->execute(this, commands[i]->tag, s); // pass param to the function
            }
            return;
        }
    }
    char str[1024];
    sprintf(str, "%s %s", cmd, s);
    if (!handleSetVariable(str, conf))
    {
        printf("Invalid variable or command name!");
    }
}


/*
 *
 */
bool tveConsole::handleSetVariable(char *args, bool conf)
{
    char *var = getCmd(&args);
    int  len = strlen(var);

    if (!len)
    {
        return false;
    }
    for (int i=0; i<var_idx; i++)
    {
        if (!strncasecmp(variables[i]->name, var, len))
        {
            if (strlen(args))
            {
                switch (variables[i]->type)
                {
                    case VT_YESNO:
                    case VT_ONOFF:
                    case VT_BOOLEAN:
                    case VT_TEXT:
                        {
                            free(variables[i]->value.text);
                            variables[i]->value.text = strdup(args);
                        }
                        break;
                    case VT_FLOAT:
                        {
                            variables[i]->value.floating = (float)atof(args);
                        }
                        break;
                    case VT_INTEGER:
                        {
                            variables[i]->value.integer = atoi(args);
                        }
                        break;
                }
                if (variables[i]->onChange && !conf)
                {
                    variables[i]->onChange();
                }
            }
            else
            {
                switch (variables[i]->type)
                {
                    case VT_YESNO:
                    case VT_ONOFF:
                    case VT_BOOLEAN:
                    case VT_TEXT:     printf("%s = %s (default: %s)",    variables[i]->name, variables[i]->value.text,     variables[i]->def); break;
                    case VT_FLOAT:    printf("%s = %4.4f (default: %s)", variables[i]->name, variables[i]->value.floating, variables[i]->def); break;
                    case VT_INTEGER:  printf("%s = %d (default: %s)",    variables[i]->name, variables[i]->value.integer,  variables[i]->def); break;
                }
            }
            return true;
        }
    }
    return false;
}


/*
 *
 */
void tveConsole::showHelp()
{
    char str[8192];
    int i;


    printf("Commands:");
    str[0] = '\0';
    for (i=0; i<cmd_idx; i++)
    {
        strcat(str, commands[i]->name);
        strcat(str, " ");
    }
    printf("%s", str);

    printf("Variables:");
    str[0] = '\0';
    for (i=0; i<var_idx; i++)
    {
        strcat(str, variables[i]->name);
        strcat(str, " ");
    }
    printf("%s", str);
}


/*
 *
 */
void tveConsole::exec(char *args)
{
    char    buffer[256];
    gzFile  f;
    bool    conf = !strcmp(args, "config.cfg");

    console->printf("executing %s...", args);
    if ( !(f=tveVoyager::openFile(args, "rt")) )
    {
        console->printf("Error: file not found.");
        return;
    }
    while (gzgets(f, buffer, sizeof(buffer)))
    {
        if ( (buffer[0] >= 'A') && (buffer[0] <= 'z') )
        {
            buffer[strlen(buffer)-1] = '\0'; // remove lf
            if (!conf)
            {
                printf("%s", buffer);
            }
            parseCommand(buffer, conf);
        }
    }
    gzclose(f);
}


/*
 *
 */
bool tveConsole::handleWaitCommand(char *args)
{
    int  len = strlen(args);

    if (atoi(args)>0)
    {
        // it's a delay to wait
        //TODO: manage that (for instance re-put the command at the end of the command queue with a lower value for args)
        return false;
    }
    if (!strncasecmp(args, "click", len))
    {
        SDL_PumpEvents();
        if (SDL_GetMouseState(NULL, NULL)&SDL_BUTTON(1))
        {
            while (SDL_GetMouseState(NULL, NULL)&SDL_BUTTON(1))
            {
                SDL_PumpEvents();
            }
            return true;
        }
        return false;
    }
    return true;
}


/*
 *
 */
bool tveConsole::genericCommand(int tag, char *args)
{
    switch (tag&0xffff)
    {
        case GC_ECHO: printf("%s", args); return true;
        case GC_SET:  if (!handleSetVariable(args, false)) printf("Invalid variable name!"); return true;
        case GC_HELP: showHelp(); return true;
        case GC_CLS:  last_buf_pos = scr_pos = 0; return true;
        case GC_EXEC: exec(args); return true;
        case GC_WAIT: return handleWaitCommand(args);
    }
    return true;
}


/*
 *
 */
void tveConsole::resize(int width, int height, int max_height)
{
    this->width      = width;
    this->height     = height;
    this->max_height = max_height;

    if (!tex_font)
    {
        tex_font = tveTexture::load("font.tex");
        for (int i=0; i<96; i++)
        {
            GLfloat inc = (12.0f/128.0f);
            GLfloat s = (i%10)*inc;
            GLfloat t = (i/10)*inc;

            lists[i] = glGenLists(1);
            glNewList(lists[i], GL_COMPILE);
            glBegin(GL_TRIANGLE_STRIP);
              glTexCoord2f(s,     t+inc);  glVertex2f(0.0f, 24.0f);
              glTexCoord2f(s,     t);      glVertex2f(0.0f, 0.0f);
              glTexCoord2f(s+inc, t+inc);  glVertex2f(24.0f, 24.0f);
              glTexCoord2f(s+inc, t);      glVertex2f(24.0f, 0.0f);
            glEnd();
            glEndList();
        }

    }
}


/*
 *
 */
void tveConsole::saveConfig()
{
    int  fd;
    char buffer[265];

    sprintf(buffer, "%s%s", tveVoyager::path, "config.cfg");
    fd = open(buffer, O_WRONLY|O_BINARY|O_TRUNC|O_CREAT, 0644);
    if (fd >= 0)
    {
        for (int i=0; i<var_idx; i++)
        {
            switch (variables[i]->type)
            {
                case VT_YESNO:
                case VT_ONOFF:
                case VT_BOOLEAN:
                case VT_TEXT:     sprintf(buffer, "%s %s\n",    variables[i]->name, variables[i]->value.text);     break;
                case VT_FLOAT:    sprintf(buffer, "%s %4.4f\n", variables[i]->name, variables[i]->value.floating); break;
                case VT_INTEGER:  sprintf(buffer, "%s %d\n",    variables[i]->name, variables[i]->value.integer);  break;
            }
            write(fd, buffer, strlen(buffer));
        }
        close(fd);
    }
}


//=====================================================================
// tveConsoleCommand
//=====================================================================

/*
 *
 */
bool genericCommand(void *obj, int tag, char *args)
{
    return ((tveConsole*)obj)->genericCommand(tag, args);
}


/*
 *
 */
tveConsoleCommand::tveConsoleCommand(char *name, int tag, tveCommandExecute f)
{
    this->name    = name;
    this->tag     = tag;
    this->execute = f;
}


//=====================================================================
// tveConsoleVariable
//=====================================================================

/*
 *
 */
tveConsoleVariable::tveConsoleVariable(char *name, tveVariableType type, char *def, tveVarOnChange onChange)
{
    this->name      = name;
    this->def       = def;
    this->type      = type;
    switch (type)
    {
        case VT_YESNO:
        case VT_ONOFF:
        case VT_BOOLEAN:
        case VT_TEXT:     value.text     = strdup(def);      break;
        case VT_FLOAT:    value.floating = (float)atof(def); break;
        case VT_INTEGER:  value.integer  = atoi(def);       break;
    }
    this->onChange  = onChange;
}


//=====================================================================
// tveConsoleCommandItem
//=====================================================================

/*
 *
 */
tveConsoleCommandItem::tveConsoleCommandItem(char *name, int tag, tveCommandExecute execute, char *args) : tveConsoleCommand(name, tag, execute)
{
    this->args = strdup(args);
}


/*
 *
 */
tveConsoleCommandItem::~tveConsoleCommandItem()
{
    delete args;
}
