/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveConsole.h
 * Creation date : 15/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#ifndef _TVECONSOLE_H_
#define _TVECONSOLE_H_

#include "voyager.h"
#include "tveLists.h"


// FIXME: allow dynamic size (for now those values should be more than enough)
const int   CON_MAX_COMMAND  = 64;
const int   CON_MAX_VARIABLE = 128;
const int   CON_SCR_BUF_SIZE = 256;
const int   CON_BUF_SIZE     = 65536;
const int   CON_HIS_SIZE     = 256;
const int   CON_MAX_LINE     = 8192;
const int   CON_CHAR_HEIGHT  = 16;
const float CON_CHAR_WIDTH   = 20.0f;


enum tveGenericCommand
{
    GC_ECHO = 1,
    GC_SET,
    GC_HELP,
    GC_CLS,
    GC_EXEC,
    GC_WAIT,
};

enum tveCommandType
{
    CT_DIRECT  = 0x000000,
    CT_DELAYED = 0x010000,
};

enum tveVariableType
{
    VT_TEXT,
    VT_INTEGER,
    VT_BOOLEAN,
    VT_FLOAT,
    VT_YESNO,
    VT_ONOFF,
};

class tveConsole;
class tveConsoleVariable;
class tveConsoleCommand;
class tveConsoleCommandQueue;
class tveConsoleCommandItem;

typedef bool (*tveCommandExecute)(void *, int, char *);
typedef void (*tveVarOnChange)(void);

typedef struct
{
    int     x;
    int     y;
    Uint32  col;
    float   scale;
    char    *text;
} tveConsoleScreenBuffer;

class tveConsoleCommand
{
public:
    tveConsoleCommand(char *name, int tag, tveCommandExecute execute);
    char *name;
    int  tag;
    tveCommandExecute execute;
};


class tveConsoleVariable
{
public:
    tveConsoleVariable(char *name, tveVariableType type, char *def, tveVarOnChange onChange);
    char    *name;
    char    *def;   // default value in string format
    tveVariableType type;
    union {
        char    *text;
        float   floating;
        int     integer;
        bool    boolean;
    } value;
    tveVarOnChange  onChange;
};


bool genericCommand(void *obj, int, char *);

class tveConsole
{
friend bool genericCommand(void *obj, int, char *);
private:
    FILE    *f;
    GLuint  tex_font;
    GLuint  lists[96];
    int     width;
    int     height;
    float   ypos; // used for animation
    int     max_height;
    int     speed;
    char    *buffer[CON_BUF_SIZE];
    char    *history[CON_HIS_SIZE];
    char    line[CON_MAX_LINE];
    int     last_his_pos;  // last position in the history buffer
    int     last_buf_pos;  // last position in the console buffer
    int     his_pos;  // current position n the history
    int     scr_pos;  // screen position in the console buffer
    int     line_pos; // position of the cursor in the editing line
    int     cmd_idx;
    tveConsoleCommand *commands[CON_MAX_COMMAND];
    int     var_idx;
    tveConsoleVariable *variables[CON_MAX_VARIABLE];
    int     scr_buf_idx;
    tveConsoleScreenBuffer screen_buffer[CON_SCR_BUF_SIZE];
    tveQueue *command_queue;

    char *getCmd(char **s);
    void parseCommand(char *s, bool conf);
    bool genericCommand(int tag, char *args);
    bool handleSetVariable(char *args, bool conf);
    bool handleWaitCommand(char *args);
    void showHelp();
    void addHistory();

public:
    tveConsole(int speed);
    ~tveConsole();
    void printf(int x, int y, Uint32 col, GLfloat scale, char *lpFmt, ...);
    void printf(int x, int y, char *lpFmt, ...);
    void printf(char *lpFmt, ...);
    void printBuffer();
    void flushBuffer();
    void display();
    bool visible();
    void handleCommands();
    void handleKeyDown(SDL_keysym keysym);
    void start();
    void stop();
    void addCommand(tveConsoleCommand *command);
    void addVariable(tveConsoleVariable *variable);
    void exec(char *);
    void resize(int width, int height, int max_height);
    void saveConfig();
};


class tveConsoleCommandItem : public tveItem, public tveConsoleCommand
{
public:
    char *args;
    tveConsoleCommandItem(char *name, int tag, tveCommandExecute execute, char *args);
    ~tveConsoleCommandItem();
};


#endif
