/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveHeightmap.cpp
 * Creation date : 23/02/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveHeightmap.h"

#define pa (*A)
#define pb (*B)
#define pc (*C)
#define pd (*D)
#define pe (*E)
#define pf (*F)
#define pg (*G)
#define ph (*H)
#define pi (*I)
#define RAND_VALUE 5

Sint16 *A, *B, *C, *D, *E, *F, *G, *H, *I;

int inline POW2(int x)
{
    int res = 1;

    while (x > res)
    {
        res *= 2;
    }
    return res;
}


/*
 *
 */
tveHeightMap::tveHeightMap()
{
    tex = new GLuint[NUM_TEX];
    char water_name[] = "water00.tex";
    for (int i=0; i<16; i++)
    {
        water_name[5] = i/10 + '0';
        water_name[6] = i%10 + '0';
        tex[WATER+i]= tveTexture::load(water_name);
    }
    tex[GRASS]  = tveTexture::load("grass.tex");
    tex[ROCK]   = tveTexture::load("rock.tex");
    tex[SNOW]   = tveTexture::load("snow.tex");
    tex[DIRT]   = tveTexture::load("dirt.tex");
    tveRGBA col = {255,0,255,0};
    tex[TREE]   = tveTexture::loada("tree.tex", col);
    tex[SHADOW] = 0; // will be loaded when we have the datas for the heightmap
    tex[SKY]    = tveTexture::generatySky();
}


/*
 *
 */
tveHeightMap::~tveHeightMap()
{
    if (tex[SHADOW])
    {
        glDeleteTextures(1, &tex[SHADOW]);
    }
    if (tex[SKY])
    {
        glDeleteTextures(1, &tex[SKY]);
    }
    delete []tex;
}


/*
 *
 */
GLuint tveHeightMap::generateShadowTex()
{
    GLuint  tex_shadow;
    tveRGBA *shadow_map = new tveRGBA[width*height*4];

    console->printf("Generating shadow texture...");
    for (int j=0; j<height; j++)
    {
        for (int i=0; i<width-1; i++)
        {
            GLubyte shadow = MIN((detail[i+1][j]*8-detail[i][j]*8)+192, 255);

            shadow_map[j*width+i][RED]   = shadow;
            shadow_map[j*width+i][GREEN] = shadow;
            shadow_map[j*width+i][BLUE]  = shadow;
            shadow_map[j*width+i][ALPHA] = 0xff;
        }
    }

    glGenTextures(1, &tex_shadow);
    glBindTexture(GL_TEXTURE_2D, tex_shadow);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, width, height, GL_RGBA, GL_UNSIGNED_BYTE, shadow_map);
    delete []shadow_map;
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);

    return tex_shadow;
}


/*
 *
 */
void tveHeightMap::smooth()
{
    int cnt;
    //smooth it and add peak //TODO: find a better solution for that
    for (cnt=0; cnt<2; cnt++)
    {
        for (int j=1; j<height; j++)
        {
            for (int i=1; i<width; i++)
            {
                if (cnt==2)
                {
                    detail[i][j] *= 16;
                }
                else
                {
                    detail[i][j] = (detail[i+1][j]+detail[i][j-1]+detail[i-1][j]+detail[i][j+1])/4;
                    if (rand()%250==1 && detail[i][j] >= 0 && cnt==0)
                    {
                        detail[i][j]    += 5;
                        detail[i-1][j]   = detail[i][j];
                        detail[i+1][j]   = detail[i][j];
                        detail[i][j-1]   = detail[i][j];
                        detail[i][j+1]   = detail[i][j];
                        detail[i-1][j-1] = detail[i][j]-3;
                        detail[i+1][j-1] = detail[i][j]-3;
                        detail[i+1][j+1] = detail[i][j]-3;
                        detail[i-1][j+1] = detail[i][j]-3;
                    }
                }
            }
        }
    }
    for (cnt=0; cnt<=width; cnt++)
    {
        detail[cnt][0]      = 1;
        detail[cnt][height] = 1;
    }
    for (cnt=0; cnt<=height; cnt++)
    {
        detail[0][cnt]     = 1;
        detail[width][cnt] = 1;
    }
}


/*
 *
 */
void tveHeightMap::generate(int height, int width, int base_y, int method)
{
    this->height = height;
    this->width  = width;
    detail[0][0]          = base_y;
    detail[width][height] = base_y;
    detail[0][height]     = base_y;
    detail[width][0]      = base_y;

    switch (method)
    {
        case 0: recurse0(0, 0, width, height); break;
        case 1: recurse1(0, 0, width, height); break;
        case 2: recurse2(0, 0, width, height); break;
    }
    tex[SHADOW] = generateShadowTex();
    generateCoarseLevel();
    generateGeoMap();
}


/*
 *
 */
void tveHeightMap::recurse0(int left, int top, int right, int bottom)
{
    int mx, my;

    my = (bottom+top)/2;
    mx = (right+left)/2;
    A = &detail[left][top];
    B = &detail[right][top];
    C = &detail[right][bottom];
    D = &detail[left][bottom];
    E = &detail[mx][my];
    F = &detail[mx][top];
    G = &detail[right][my];
    H = &detail[mx][bottom];
    I = &detail[left][my];
    pe = (pa + pb + pc + pd) / 4 + SRANDOM(RAND_VALUE);
    pf = (pe + pa + pb) / 3 + SRANDOM(RAND_VALUE);
    pg = (pe + pb + pc) / 3 + SRANDOM(RAND_VALUE);
    ph = (pe + pc + pd) / 3 + SRANDOM(RAND_VALUE);
    pi = (pe + pa + pd) / 3 + SRANDOM(RAND_VALUE);

    if ((bottom-top > 1) && (right-left > 1))
    {
        recurse0(left, top, mx, my);
        recurse0(mx, top, right, my);
        recurse0(left, my, mx, bottom);
        recurse0(mx, my, right, bottom);
    }
}


/*
 *
 */
void tveHeightMap::recurse1(int left, int top, int right, int bottom)
{
    recurse2(left, bottom, right, top);
}


/*
 *
 */
//FIXME: this is not 100% correct
void tveHeightMap::recurse2(int left, int bottom, int right, int top)
{
    int mx, my;

    my = (bottom+top)/2;
    mx = (right+left)/2;
    A = &detail[left][top];
    B = &detail[left][bottom];
    C = &detail[right][top];
    D = &detail[right][bottom];
    E = &detail[mx][my];
    F = &detail[mx][top];
    //G = &detail[right][my];
    //H = &detail[mx][bottom];
    I = &detail[left][my];
    pe = (pa + pb + pc + pd) / 4;
    if ((top-bottom < height) && (top-bottom > 2))
    {
        pe += SRANDOM(top-bottom);
    }
    pf = (pa + pc + pe) / 3;
    //pg = (pc + pd + pe) / 3;
    //ph = (pd + pb + pe) / 3;
    pi = (pb + pa + pe) / 3;
    /*
    if ((top-bottom < height) && (top-bottom > 2))
    {
        pe += SRANDOM(top-bottom);
        pf += SRANDOM((top-bottom)/2);
        //pg += SRANDOM((top-bottom)/2);
        //ph += SRANDOM((top-bottom)/2);
        pi += SRANDOM((top-bottom)/2);
    }
    */
    if ((top-bottom > 2) && (right-left > 2))
    {
        recurse2(left, my, mx, top);
        recurse2(left, bottom, mx, my);
        recurse2(mx, my, right, top);
        recurse2(mx, bottom, right, my);
    }
}


/*
 *
 */
//TODO: manage endianess
void tveHeightMap::loadFromFile(char *filename)
{
    char    file_name[256];
    gzFile  f;

    sprintf(file_name, "../data/heightmaps/%s", filename);
    if ( (f=tveVoyager::openFile(file_name, "rb")) )
    {
        char ext[4];

        console->printf("loading... %-24s ", filename);
        for (int i=0; i<4; i++)
        {
            ext[i] = filename[strlen(filename)-3+i];
        }
        memset(detail, 0, (MAXX+1)*(MAXZ+1));
        //FIXME: do format recognition
        if (!strcasecmp(ext, "ter"))
        {
            loadFromTGN(f);
        }
        else if (!strcasecmp(ext, "tve"))
        {
            loadFromTGA(f);
        }
        else if (!strcasecmp(ext, "h16"))
        {
            loadFromH16(f);
        }
        gzclose(f);
        tex[SHADOW] = generateShadowTex();
        generateCoarseLevel();
        generateGeoMap();
    }
    else
    {
        console->printf("cannot load file %s ", filename);
    }
}


typedef struct
{
    Sint16 s1;
    Sint16 s2;
} tveTwoSint16;

/*
 *
 */
// file format description: http://www.planetside.co.uk/terragen/dev/tgterrain.html
void tveHeightMap::loadFromTGN(gzFile f)
{
    char marker[20];

    gzread(f, marker, 16);
    marker[16] = '\0';
    if (strcmp(marker, "TERRAGENTERRAIN "))
    {
        width = height = 16; //FIXME: this is not the best thing to do
        return;              // if it's not a valid TerraGen terrain
    }
    gzread(f, marker, 4);  // read first marker
    marker[4] = '\0';
    while (true) //FIXME: we hope it's a valid file ;)
    {
        if (!strcmp(marker, "XPTS"))
        {
            tveTwoSint16 ts;
            gzread(f, &ts, sizeof(ts));
            width = SDL_SwapLE16(ts.s1);
        }
        else if (!strcmp(marker, "YPTS"))
        {
            tveTwoSint16 ts;
            gzread(f, &ts, sizeof(ts));
            height = SDL_SwapLE16(ts.s1);
        }
        else if (!strcmp(marker, "SIZE"))
        {
            tveTwoSint16 ts;
            gzread(f, &ts, sizeof(ts));
            height = width = SDL_SwapLE16(ts.s1)+1;
        }
        else if (!strcmp(marker, "SCAL")) // we ignore it
        {
            tveVertex vertex;
            gzread(f, &vertex, sizeof(vertex));
        }
        else if (!strcmp(marker, "CRAD")) // we ignore it
        {
            float fl;
            gzread(f, &fl, sizeof(fl));
        }
        else if (!strcmp(marker, "CRVM")) // we ignore it
        {
            Uint32 i;
            gzread(f, &i, sizeof(i));
        }
        else if (!strcmp(marker, "ALTW"))
        {
            tveTwoSint16 ts;
            Sint16       *buffer  ;

            gzread(f, &ts, sizeof(ts));
            ts.s1 = SDL_SwapLE16(ts.s1);
            ts.s2 = SDL_SwapLE16(ts.s2);
            buffer = new Sint16[height*width];
            gzread(f, buffer, height*width*2);
            console->printf(" | w: %d  h: %d  bias: %d  scale: %d", width, height, ts.s2, ts.s1);
            for (int j=0; j<height; j++)
            {
                for (int i=0; i<height; i++)
                {
                    detail[i][j] = 128+ts.s2 + buffer[j*width+i]*ts.s1/16384;//8912;//16384;//65536;
                }
            }
            delete []buffer;
            if ((height*width) % 2)
            {
                gzread(f, marker, 2);
            }
            break;
        }
        gzread(f, marker, 4);  // read next marker
        marker[4] = '\0';
    }
    saveTGA();
}


/*
 * X-Isle format
 */
void tveHeightMap::loadFromH16(gzFile f)
{
    Uint16 *buffer = new Uint16[1024*1024];

    gzread(f, buffer, 1024*1024*2);
    height = 1024;
    width = 1024;
    for (int j=0; j<height; j++)
    {
        for (int i=0; i<height; i++)
        {
            detail[i][j] = buffer[j*width+i]/32-128;
        }
    }
    console->printf(" | w: %d  h: %d", width, height);
}


/*
 *
 */
void tveHeightMap::loadFromTGA(gzFile f)
{
    tveTGAHeader tga_hdr;
    char         *field;

    gzread(f, &tga_hdr, sizeof(tveTGAHeader));
    tga_hdr.width  = SDL_SwapLE16(tga_hdr.width);
    tga_hdr.height = SDL_SwapLE16(tga_hdr.height);
    this->width  = POW2(tga_hdr.width);
    this->height = POW2(tga_hdr.height);
    console->printf(" | w: %d  h: %d -> w: %d  h: %d",
                    tga_hdr.width, tga_hdr.height, this->width, this->height);
    field = new char[tga_hdr.id_field_length];
    gzread(f, field, tga_hdr.id_field_length); // ignore TGA informations
    delete []field;

    int x_offset = (this->width-tga_hdr.width)/2;
    int z_offset = (this->height-tga_hdr.height)/2;

    if (tga_hdr.image_pix_size==8)
    {
        Uint8 *image = new Uint8[tga_hdr.width*tga_hdr.height];
        gzread(f, image, tga_hdr.width * tga_hdr.height);
        for  (int j=0; j<tga_hdr.height; j++)
        {
            for  (int i=0; i<tga_hdr.width; i++)
            {
                detail[i+x_offset][j+z_offset] = image[j*tga_hdr.width+i];
            }
        }
        delete []image;
    }
    else
    {
        tveRGB *image = new tveRGB[tga_hdr.width*tga_hdr.height];
        gzread(f, image, (long)tga_hdr.width * (long)tga_hdr.height * (long)sizeof(tveRGB));
        for  (int j=0; j<tga_hdr.height; j++)
        {
            for  (int i=0; i<tga_hdr.width; i++)
            {
                detail[i+x_offset][j+z_offset] = image[j*tga_hdr.width+i][0];
            }
        }
        delete []image;
    }
}


/*
 *
 */
void tveHeightMap::generateGeoMap()
{
    memset(geomap, 0, (height+1)*(width+1));
    for (int j=0; j<height; j++)
    {
        for (int i=0; i<width; i++)
        {
            if ( (detail[i][j] > tveVoyager::cv_water->value.floating/SCALE_Y) &&
                 (detail[i][j] < 128) &&
                 (detail[i][j]-detail[i+1][j] <  4) &&
                 (detail[i][j]-detail[i+1][j] > -4) &&
                 (detail[i][j]-detail[i][j+1] <  4) &&
                 (detail[i][j]-detail[i][j+1] > -4) &&
                 (!RANDOM(6)) )
            {
                geomap[i][j] |= GEO_TREE;
            }
        }
    }
}


/*
 *
 */
void tveHeightMap::generateCoarseLevel()
{
    int j;

    for (j=1; j<height/2; j++)
    {
        for (int i=1; i<width/2; i++)
        {
            int height = detail[2*i][2*j]*4;
            height += detail[2*i-1][2*j+0]*2;
            height += detail[2*i+1][2*j+0]*2;
            height += detail[2*i+0][2*j-1]*2;
            height += detail[2*i+0][2*j-1]*2;
            height += detail[2*i-1][2*j-1];
            height += detail[2*i+1][2*j-1];
            height += detail[2*i-1][2*j+1];
            height += detail[2*i+1][2*j+1];
            coarse[i][j] = height/16;
            coarse[i][j] = detail[2*i][2*j];
        }
    }
    for (j=0; j<height/2; j++)
    {
        coarse[0][j]       = detail[0][2*j];
        coarse[width/2][j] = detail[width][2*j];
    }
    for (j=0; j<width/2; j++)
    {
        coarse[j][0]        = detail[2*j][0];
        coarse[j][height/2] = detail[2*j][height];
    }
}


/*
 *
 */
void tveHeightMap::saveTGA()
{
    int          fd;
    tveTGAHeader tga_hdr;
    int          size = (height-1)*(width-1);
    Uint8        *buffer = new Uint8[size];

    if (sizeof(tveTGAHeader) != 18) 
    {
        printf("Erreur: saveTGA(): sizeof(tveTGAHeader) = %u\n", sizeof(tveTGAHeader));
        exit(-1);
    }
    memset(&tga_hdr, 0, sizeof(tga_hdr));
    tga_hdr.width = width-1;
    tga_hdr.height = height-1;
    tga_hdr.image_pix_size = 8;
    tga_hdr.image_type = 2;
    tga_hdr.image_descriptor = 32;
    for (int j=0; j<height-1; j++)
    {
        for (int i=0; i<width-1; i++)
        {
            buffer[i+j*(width-1)] = MIN( MAX(detail[i][j], 0), 255 );
        }
    }

    char tganame[] = "../data/heightmaps/terrain00.tga";
    char file_name[256];
    for (int i=0; i<100; i++)
    {
        tganame[26] = i/10  + '0';
        tganame[27] = i%10  + '0';
        if (!tveVoyager::fileExists(tganame))
        {
            break;
        }
    }
    console->printf("saving... %s\n", tganame);
    sprintf(file_name, "%s%s", tveVoyager::path, tganame);
    fd = open(file_name, O_WRONLY|O_BINARY|O_CREAT, 0644);
    if (fd < 0)
    {
        console->printf("cannot save to %s ", tganame);
        return;
    }
    write(fd, &tga_hdr, sizeof(tveTGAHeader));
    write(fd, buffer, size);
    close(fd);
}
