/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveModel.cpp
 * Creation date : 21/04/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveTexture.h"
#include "tveModel.h"


typedef struct
{
   Uint8 vertex[3];
   Uint8 light_normal_index;
} tveMdzAPolygonVertex;


typedef struct
{
   float scale[3];
   float translate[3];
   char  name[16];
   tveMdzAPolygonVertex alias_vertices;
} tveMdzAFrame;



/*
 *
 */
tveGenericModel::tveGenericModel()
{
    tex = 0;
}


/*
 *
 */
tveGenericModel::~tveGenericModel()
{
    if (tex)
    {
        glDeleteTextures(1, &tex);
    }
}


/*
 *
 */
tveMdzModel::tveMdzModel(gzFile f)
{
    memset(&header, 0, sizeof(header));
    skins  = NULL;
    frames = NULL;
    tex_coords = NULL;
    polygons = NULL;
    gl_command_buffer = NULL;

    gzread(f, &header, sizeof(tveMdzHeader));
    for (unsigned i=0, *buffer=(unsigned*)&header; i<sizeof(tveMdzHeader)/sizeof(int); i++)
    {
        buffer[i] = SDL_SwapLE32(buffer[i]);
    }

    // test if the file looks like an MD2
    if (header.magic != ('2'<<24) + ('P'<<16) + ('D'<<8) + 'I')
    {
        throw NULL;
        return;
    }
    // read skins names
    if (header.num_skins > 0)
    {
        skins = new tveMdzSkin[header.num_skins];
        gzseek(f, header.offset_skins, SEEK_SET);
        for (int i=0; i<header.num_skins; i++)
        {
            gzread(f, &skins[i], sizeof(tveMdzSkin));
        }
    }
    // read texture coordinates
    if (header.num_tex_coords > 0)
    {
        tex_coords = new tveMdzTexCoord[header.num_tex_coords];
        gzseek(f, header.offset_tex_coords, SEEK_SET);
        for (int i=0; i<header.num_tex_coords; i++)
        {
            gzread(f, &tex_coords[i], sizeof(tveMdzTexCoord));
            tex_coords[i].s = SDL_SwapLE16(tex_coords[i].s);
            tex_coords[i].t = SDL_SwapLE16(tex_coords[i].t);
        }
    }
    // read polygons
    if (header.num_polygons > 0)
    {
        polygons = new tveMdzPolygons[header.num_polygons];
        gzseek(f, header.offset_polygons, SEEK_SET);
        for (int i=0; i<header.num_polygons; i++)
        {
            gzread(f, &polygons[i], sizeof(tveMdzPolygons));
            for (int j=0; j<3; j++)
            {
                polygons[i].vertex_indices[j]  = SDL_SwapLE16(polygons[i].vertex_indices[j]);
                polygons[i].texture_indices[j] = SDL_SwapLE16(polygons[i].texture_indices[j]);
            }
        }
    }
    // read frames
    if (header.num_frames > 0)
    {
        frames = new tveMdzFrame[header.num_frames];
        gzseek(f, header.offset_frames, SEEK_SET);
        for (int i=0; i<header.num_frames; i++)
        {
            int                  j;
            Uint8                *buffer = new Uint8[header.frame_size];
            tveMdzAFrame         *frame = (tveMdzAFrame *) buffer;
            tveMdzAPolygonVertex *alias_vertices = &frame->alias_vertices;

            gzread(f, frame, header.frame_size);
            strcpy(frames[i].name, frame->name);
            for (j=0; j<3; j++)
            {
                int scale     = SDL_SwapLE32(*((int *)&frame->scale[j]));
                int translate = SDL_SwapLE32(*((int *)&frame->translate[j]));

                frame->scale[j]     = *((float *)&scale);
                frame->translate[j] = *((float *)&translate);
            }
            frames[i].vertices = new tveMdzPolygonsVertex[header.num_vertices];
            for (j=0; j<header.num_vertices; j++)
            {
                frames[i].vertices[j].vertex[0] = alias_vertices[j].vertex[0] * frame->scale[0] + frame->translate[0];
                frames[i].vertices[j].vertex[2] = -1 * (alias_vertices[j].vertex[1] * frame->scale[1] + frame->translate[1]);
                frames[i].vertices[j].vertex[1] = alias_vertices[j].vertex[2] * frame->scale[2] + frame->translate[2];
            }
        }
    }
    // read gl commands
    if (header.num_gl_commands)
    {
        gl_command_buffer = new int[header.num_gl_commands];
        gzseek(f, header.offset_gl_commands, SEEK_SET);
        gzread(f, gl_command_buffer, sizeof(Sint32)*header.num_gl_commands);
        for (int j=0; j<header.num_gl_commands; j++)
        {
            gl_command_buffer[j] = SDL_SwapLE32(gl_command_buffer[j]);
        }
    }
    loadSkin(f);
    frame_num = 0;
}


/*
 *
 */
tveMdzModel::~tveMdzModel()
{
    if (skins)
    {
        delete []skins;
    }
    if (frames)
    {
        for (int i=0; i<header.num_frames; i++)
        {
            if (frames[i].vertices)
            {
                delete []frames[i].vertices;
            }
        }
        delete []frames;
    }
    if (tex_coords)
    {
        delete []tex_coords;
    }
    if (polygons)
    {
        delete []polygons;
    }
    if (gl_command_buffer)
    {
        delete []gl_command_buffer;
    }
}


/*
 *
 */
void tveMdzModel::loadSkin(gzFile f)
{
    tvePCXHeader pcx_header;
    tveRGB  palette[256];
    tveRGB  *image;
    int     w, h;
    int     ptr = 0;
    int     ch, rep;
    int     size;

    gzseek(f, header.offset_end, SEEK_SET);
    if (gzread(f, &pcx_header, sizeof(tvePCXHeader)) < 0)
    {
        gzclose(f);
        return;
    }
    pcx_header.xmax  = SDL_SwapLE16(pcx_header.xmax);
    pcx_header.ymax  = SDL_SwapLE16(pcx_header.ymax);
    pcx_header.xmin  = SDL_SwapLE16(pcx_header.xmin);
    pcx_header.ymin  = SDL_SwapLE16(pcx_header.ymin);

    //gzseek(f, -768, SEEK_END);
    //FIXME
    while (!gzeof(f))
    {
        ch = gzgetc(f);
    }
    gzseek(f, -768, SEEK_CUR);

    if (gzread(f, palette, 768) < 0)
    {
        gzclose(f);
        return;
    }
    w = pcx_header.xmax - pcx_header.xmin + 1;
    h = pcx_header.ymax - pcx_header.ymin + 1;
    size = w * h;
    image = new tveRGB[size];

    gzseek(f, sizeof(tvePCXHeader)+header.offset_end, SEEK_SET);
    while (ptr < size)
    {
        ch = gzgetc(f);
        if (ch >= 192)
        {
            rep = ch - 192;
            ch = gzgetc(f);
        }
        else
        {
            rep = 1;
        }
        while (rep--)
        {
            image[ptr][RED]   = palette[ch][RED];
            image[ptr][GREEN] = palette[ch][GREEN];
            image[ptr][BLUE]  = palette[ch][BLUE];
            ptr++;
        }
    }

    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_2D, tex);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, w, h, GL_RGB, GL_UNSIGNED_BYTE, image);
    delete []image;
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_NEAREST);
}


/*
 *
 */
void tveMdzModel::print()
{
    console->printf("version:         %d", header.version);
    console->printf("skin_width:      %d", header.skin_width);
    console->printf("skin_height:     %d", header.skin_height);
    console->printf("frame_size:      %d", header.frame_size);
    console->printf("num_skins:       %d", header.num_skins);
    console->printf("num_vertices:    %d", header.num_vertices);
    console->printf("num_tex_coords:  %d", header.num_tex_coords);
    console->printf("num_polygons:    %d", header.num_polygons);
    console->printf("num_gl_commands: %d", header.num_gl_commands);
    console->printf("num_frames:      %d", header.num_frames);
    console->printf("off_skins:       %d", header.offset_skins);
    console->printf("off_tex_coords:  %d", header.offset_tex_coords);
    console->printf("off_polygons:    %d", header.offset_polygons);
    console->printf("off_frames:      %d", header.offset_frames);
    console->printf("off_gl_commands: %d", header.offset_gl_commands);
    console->printf("off_end:         %d", header.offset_end);
    /*
    for (int i=0; i<model->header.num_frames; i++)
    {
        console->printf("%s ", model->frames[i].name);
    }
    */
}


/*
 *
 */
void tveMdzModel::render()
{
    tveMdzFrame *frame = &frames[(int)frame_num];
    int         *gl_cmd_buffer = gl_command_buffer;
    int         val, count, index;
    GLfloat     s, t;

    frame_num += tveVoyager::ratio*0.01f;
    if (frame_num > header.num_frames)
    {
        frame_num -= header.num_frames;
    }

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, tex);

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    //TODO: put that in a display list?
    glPushMatrix(); 
    glTranslatef(256, 16, 256);
    //glRotatef(pos->angley, 0.0f, 1.0f, 0.0f);
    glScalef(0.5f, 0.5f, 0.5f);

    val = *gl_cmd_buffer++;
    while (val != 0)
    {
        if (val < 0)
        {
            glBegin(GL_TRIANGLE_FAN);
            count = -val;
        }
        else
        {
            glBegin(GL_TRIANGLE_STRIP);
            count = val;
        }
        while (count--)
        {
            s = *(float *) gl_cmd_buffer++;
            t = *(float *) gl_cmd_buffer++;
            index = *gl_cmd_buffer++;
            glTexCoord2f(s, t);
            glVertex3f(frame->vertices[index].vertex[0]/2.0f,
                       frame->vertices[index].vertex[1]/2.0f,
                       frame->vertices[index].vertex[2]/2.0f);
        }
        glEnd ();
        val = *gl_cmd_buffer++;
    }
    glPopMatrix();
}


/*
 *
 */
tveMS3DModel::tveMS3DModel(gzFile f)
{
//TODO
}


/*
 *
 */
tveMS3DModel::~tveMS3DModel()
{
//TODO
}


/*
 *
 */
void tveMS3DModel::print()
{
//TODO
}


/*
 *
 */
void tveMS3DModel::render()
{
//TODO
}


/*
Model::Model()
{
        m_numMeshes = 0;
        m_pMeshes = NULL;
        m_numMaterials = 0;
        m_pMaterials = NULL;
        m_numTriangles = 0;
        m_pTriangles = NULL;
        m_numVertices = 0;
        m_pVertices = NULL;
}

Model::~Model()
{
        int i;
        for ( i = 0; i < m_numMeshes; i++ )
                delete[] m_pMeshes[i].m_pTriangleIndices;
        for ( i = 0; i < m_numMaterials; i++ )
                delete[] m_pMaterials[i].m_pTextureFilename;

        m_numMeshes = 0;
        if ( m_pMeshes != NULL )
        {
                delete[] m_pMeshes;
                m_pMeshes = NULL;
        }

        m_numMaterials = 0;
        if ( m_pMaterials != NULL )
        {
                delete[] m_pMaterials;
                m_pMaterials = NULL;
        }

        m_numTriangles = 0;
        if ( m_pTriangles != NULL )
        {
                delete[] m_pTriangles;
                m_pTriangles = NULL;
        }

        m_numVertices = 0;
        if ( m_pVertices != NULL )
        {
                delete[] m_pVertices;
                m_pVertices = NULL;
        }
}

void Model::draw() 
{
        GLboolean texEnabled = glIsEnabled( GL_TEXTURE_2D );

        // Draw by group
        for ( int i = 0; i < m_numMeshes; i++ )
        {
                int materialIndex = m_pMeshes[i].m_materialIndex;
                if ( materialIndex >= 0 )
                {
                        glMaterialfv( GL_FRONT, GL_AMBIENT, m_pMaterials[materialIndex].m_ambient );
                        glMaterialfv( GL_FRONT, GL_DIFFUSE, m_pMaterials[materialIndex].m_diffuse );
                        glMaterialfv( GL_FRONT, GL_SPECULAR, m_pMaterials[materialIndex].m_specular );
                        glMaterialfv( GL_FRONT, GL_EMISSION, m_pMaterials[materialIndex].m_emissive );
                        glMaterialf( GL_FRONT, GL_SHININESS, m_pMaterials[materialIndex].m_shininess );

                        if ( m_pMaterials[materialIndex].m_texture > 0 )
                        {
                                glBindTexture( GL_TEXTURE_2D, m_pMaterials[materialIndex].m_texture );
                                glEnable( GL_TEXTURE_2D );
                        }
                        else
                                glDisable( GL_TEXTURE_2D );
                }
                else
                {
                        // Material properties?
                        glDisable( GL_TEXTURE_2D );
                }

                glBegin( GL_TRIANGLES );
                {
                        for ( int j = 0; j < m_pMeshes[i].m_numTriangles; j++ )
                        {
                                int triangleIndex = m_pMeshes[i].m_pTriangleIndices[j];
                                const Triangle* pTri = &m_pTriangles[triangleIndex];

                                for ( int k = 0; k < 3; k++ )
                                {
                                        int index = pTri->m_vertexIndices[k];

                                        glNormal3fv( pTri->m_vertexNormals[k] );
                                        glTexCoord2f( pTri->m_s[k], pTri->m_t[k] );
                                        glVertex3fv( m_pVertices[index].m_location );
                                }
                        }
                }
                glEnd();
        }

        if ( texEnabled )
                glEnable( GL_TEXTURE_2D );
        else
                glDisable( GL_TEXTURE_2D );
}

void Model::reloadTextures()
{
        for ( int i = 0; i < m_numMaterials; i++ )
                if ( strlen( m_pMaterials[i].m_pTextureFilename ) > 0 )
                        m_pMaterials[i].m_texture = LoadGLTexture( m_pMaterials[i].m_pTextureFilename );
                else
                        m_pMaterials[i].m_texture = 0;
}

// byte-align structures
#ifdef _MSC_VER
#       pragma pack( push, packing )
#       pragma pack( 1 )
#       define PACK_STRUCT
#elif defined( __GNUC__ )
#       define PACK_STRUCT      __attribute__((packed))
#else
#       error you must byte-align these structures with the appropriate compiler directives
#endif

typedef unsigned char byte;
typedef unsigned short word;

// File header
struct MS3DHeader
{
        char m_ID[10];
        int m_version;
} PACK_STRUCT;

// Vertex information
struct MS3DVertex
{
        byte m_flags;
        float m_vertex[3];
        char m_boneID;
        byte m_refCount;
} PACK_STRUCT;

// Triangle information
struct MS3DTriangle
{
        word m_flags;
        word m_vertexIndices[3];
        float m_vertexNormals[3][3];
        float m_s[3], m_t[3];
        byte m_smoothingGroup;
        byte m_groupIndex;
} PACK_STRUCT;

// Material information
struct MS3DMaterial
{
    char m_name[32];
    float m_ambient[4];
    float m_diffuse[4];
    float m_specular[4];
    float m_emissive[4];
    float m_shininess;  // 0.0f - 128.0f
    float m_transparency;       // 0.0f - 1.0f
    byte m_mode;        // 0, 1, 2 is unused now
    char m_texture[128];
    char m_alphamap[128];
} PACK_STRUCT;

//      Joint information
struct MS3DJoint
{
        byte m_flags;
        char m_name[32];
        char m_parentName[32];
        float m_rotation[3];
        float m_translation[3];
        word m_numRotationKeyframes;
        word m_numTranslationKeyframes;
} PACK_STRUCT;

// Keyframe data
struct MS3DKeyframe
{
        float m_time;
        float m_parameter[3];
} PACK_STRUCT;

// Default alignment
#ifdef _MSC_VER
#       pragma pack( pop, packing )
#endif

#undef PACK_STRUCT

bool MilkshapeModel::loadModelData( const char *filename )
{
        ifstream inputFile( filename, ios::in | ios::binary | ios::nocreate );
        if ( inputFile.fail())
                return false;   // "Couldn't open the model file."

        inputFile.seekg( 0, ios::end );
        long fileSize = inputFile.tellg();
        inputFile.seekg( 0, ios::beg );

        byte *pBuffer = new byte[fileSize];
        inputFile.read( pBuffer, fileSize );
        inputFile.close();

        const byte *pPtr = pBuffer;
        MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
        pPtr += sizeof( MS3DHeader );

        if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
                return false; // "Not a valid Milkshape3D model file."

        if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
                return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

        int nVertices = *( word* )pPtr; 
        m_numVertices = nVertices;
        m_pVertices = new Vertex[nVertices];
        pPtr += sizeof( word );

        int i;
        for ( i = 0; i < nVertices; i++ )
        {
                MS3DVertex *pVertex = ( MS3DVertex* )pPtr;
                m_pVertices[i].m_boneID = pVertex->m_boneID;
                memcpy( m_pVertices[i].m_location, pVertex->m_vertex, sizeof( float )*3 );
                pPtr += sizeof( MS3DVertex );
        }

        int nTriangles = *( word* )pPtr;
        m_numTriangles = nTriangles;
        m_pTriangles = new Triangle[nTriangles];
        pPtr += sizeof( word );

        for ( i = 0; i < nTriangles; i++ )
        {
                MS3DTriangle *pTriangle = ( MS3DTriangle* )pPtr;
                int vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
                float t[3] = { 1.0f-pTriangle->m_t[0], 1.0f-pTriangle->m_t[1], 1.0f-pTriangle->m_t[2] };
                memcpy( m_pTriangles[i].m_vertexNormals, pTriangle->m_vertexNormals, sizeof( float )*3*3 );
                memcpy( m_pTriangles[i].m_s, pTriangle->m_s, sizeof( float )*3 );
                memcpy( m_pTriangles[i].m_t, t, sizeof( float )*3 );
                memcpy( m_pTriangles[i].m_vertexIndices, vertexIndices, sizeof( int )*3 );
                pPtr += sizeof( MS3DTriangle );
        }

        int nGroups = *( word* )pPtr;
        m_numMeshes = nGroups;
        m_pMeshes = new Mesh[nGroups];
        pPtr += sizeof( word );
        for ( i = 0; i < nGroups; i++ )
        {
                pPtr += sizeof( byte ); // flags
                pPtr += 32;                             // name

                word nTriangles = *( word* )pPtr;
                pPtr += sizeof( word );
                int *pTriangleIndices = new int[nTriangles];
                for ( int j = 0; j < nTriangles; j++ )
                {
                        pTriangleIndices[j] = *( word* )pPtr;
                        pPtr += sizeof( word );
                }

                char materialIndex = *( char* )pPtr;
                pPtr += sizeof( char );
        
                m_pMeshes[i].m_materialIndex = materialIndex;
                m_pMeshes[i].m_numTriangles = nTriangles;
                m_pMeshes[i].m_pTriangleIndices = pTriangleIndices;
        }

        int nMaterials = *( word* )pPtr;
        m_numMaterials = nMaterials;
        m_pMaterials = new Material[nMaterials];
        pPtr += sizeof( word );
        for ( i = 0; i < nMaterials; i++ )
        {
                MS3DMaterial *pMaterial = ( MS3DMaterial* )pPtr;
                memcpy( m_pMaterials[i].m_ambient, pMaterial->m_ambient, sizeof( float )*4 );
                memcpy( m_pMaterials[i].m_diffuse, pMaterial->m_diffuse, sizeof( float )*4 );
                memcpy( m_pMaterials[i].m_specular, pMaterial->m_specular, sizeof( float )*4 );
                memcpy( m_pMaterials[i].m_emissive, pMaterial->m_emissive, sizeof( float )*4 );
                m_pMaterials[i].m_shininess = pMaterial->m_shininess;
                m_pMaterials[i].m_pTextureFilename = new char[strlen( pMaterial->m_texture )+1];
                strcpy( m_pMaterials[i].m_pTextureFilename, pMaterial->m_texture );
                pPtr += sizeof( MS3DMaterial );
        }

        reloadTextures();

        delete[] pBuffer;

        return true;
}
*/


/*
 *
 */
tveModel::tveModel(char *filename)
{
    gzFile  f;
    char    file_name[256];
    char    ext[4];

    sprintf(file_name, "../data/models/%s", filename);
    if ( !(f=tveVoyager::openFile(file_name, "rb")) )
    {
        console->printf("Error: couldn't load %s", filename);
        model = NULL;
        return;
    }
    for (int i=0; i<4; i++)
    {
        ext[i] = filename[strlen(filename)-3+i];
    }
    //FIXME: do format recognition
    if (!strcasecmp(ext, "mdz"))
    {
        try
        {
            model = new tveMdzModel(f);
        }
        catch (...)
        {
            model = NULL; //just to be sure
        }
    }
    else if (!strcasecmp(ext, "msz"))
    {
        model = new tveMS3DModel(f);
    }
    gzclose(f);
}


/*
 *
 */
tveModel::~tveModel()
{
    if (model)
    {
        delete model;
    }
}


#if 0
/*
 *
 */
tveVertex *tveModel::getBoundingBox()
{
    if (!model)
    {
        return NULL;
    }
    static tveVertex minmax[2];

    minmax[0].x = minmax[0].y = minmax[0].z = 999999.0f;
    minmax[1].x = minmax[1].y = minmax[1].z = -999999.0f;

    // get bounding box
    for (int i = 0; i < header.num_vertices; i++)
    {
        tveMdzPolygonsVertex *v = &frames[0].vertices[i];

        if (v->vertex[0] < minmax[0].x)
        {
            minmax[0].x = v->vertex[0];
        }
        else if (v->vertex[0] > minmax[1].x)
        {
            minmax[1].x = v->vertex[0];
        }
        if (v->vertex[1] < minmax[0].y)
        {
            minmax[0].y = v->vertex[1];
        }
        else if (v->vertex[1] > minmax[1].y)
        {
            minmax[1].y = v->vertex[1];
        }
        if (v->vertex[2] < minmax[0].z)
        {
            minmax[0].z = v->vertex[2];
        }
        else if (v->vertex[2] > minmax[1].z)
        {
            minmax[1].z = v->vertex[2];
        }
    }
    return minmax;
}


/*
 *
 */
int tveModel::getAnimationCount()
{
    if (!model)
    {
        return 0;
    }
    int i, j, pos;
    int count;
    char name[16], last[16];

    strcpy (last, frames[0].name);
    pos = strlen (last) - 1;
    j = 0;
    while (last[pos] >= '0' && last[pos] <= '9' && j < 2)
    {
        pos--;
        j++;
    }
    last[pos + 1] = '\0';

    count = 0;

    for (i=0; i<=header.num_frames; i++)
    {
        if (i==header.num_frames)
        {
            name[0] = '\0';
        }
        else
        {
            strcpy (name, frames[i].name);
        }
        pos = strlen (name) - 1;
        j = 0;
        while (name[pos] >= '0' && name[pos] <= '9' && j < 2)
        {
            pos--;
            j++;
        }
        name[pos + 1] = '\0';

        if (strcmp (last, name))
        {
            strcpy (last, name);
            count++;
        }
    }

    return count;
}


/*
 *
 */
char *tveModel::getAnimationName(int animation)
{
    if (!model)
    {
        return NULL;
    }
    int i, j, pos;
    int count;
    static char last[32];
    char name[32];

    strcpy (last, frames[0].name);
    pos = strlen (last) - 1;
    j = 0;
    while (last[pos] >= '0' && last[pos] <= '9' && j < 2)
    {
        pos--;
        j++;
    }
    last[pos + 1] = '\0';

    count = 0;

    for (i=0; i<=header.num_frames; i++)
    {
        if (i==header.num_frames)
        {
            name[0] = '\0';
        }
        else
        {
            strcpy (name, frames[i].name);
        }
        pos = strlen (name) - 1;
        j = 0;
        while (name[pos] >= '0' && name[pos] <= '9' && j < 2)
        {
            pos--;
            j++;
        }
        name[pos + 1] = '\0';

        if (strcmp (last, name))
        {
            if (count == animation)
            {
                return last;
            }
            strcpy (last, name);
            count++;
        }
    }

    return 0;
}


/*
 *
 */
void tveModel::getAnimationFrames(int animation, int *startFrame, int *endFrame)
{
    if (!model)
    {
        return;
    }
    int i, j, pos;
    int count, num_frames, frameCount;
    char name[16], last[16];

    strcpy (last, frames[0].name);
    pos = strlen (last) - 1;
    j = 0;
    while (last[pos] >= '0' && last[pos] <= '9' && j < 2)
    {
        pos--;
        j++;
    }
    last[pos + 1] = '\0';

    count = 0;
    num_frames = 0;
    frameCount = 0;

    for (i = 0; i <= header.num_frames; i++)
    {
        if (i == header.num_frames)
        {
            name[0] = '\0';
        }
        else
        {
            strcpy (name, frames[i].name);
        }
        pos = strlen (name) - 1;
        j = 0;
        while (name[pos] >= '0' && name[pos] <= '9' && j < 2)
        {
            pos--;
            j++;
        }
        name[pos + 1] = '\0';

        if (strcmp (last, name))
        {
            strcpy (last, name);

            if (count == animation)
            {
                *startFrame = frameCount - num_frames;
                *endFrame = frameCount - 1;
                return;
            }

            count++;
            num_frames = 0;
        }
        frameCount++;
        num_frames++;
    }
    *startFrame = *endFrame = 0;
}
#endif


/*
 *
 */
void tveModel::print()
{
    if (!model)
    {
        console->printf("Warning: model not loaded.");
        return;
    }
    model->print();
}


/*
 *
 */
void tveModel::render()
{
    if (model)
    {
        model->render();
    }
}
