/*
 * Copyright (c) 2001, Thierry Van Elsuw 
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of DS Improve nor the names of its contributors may 
 * be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * Module name   : tveSector.cpp
 * Creation date : 26/04/2001
 * e-mail        : Thierry Van Elsuwe <tvelsuwe@swing.be>
 */

#include "voyager.h"
#include "tveSector.h"


//=====================================================================
// tveQuadSector
//=====================================================================

/*
 *
 */
tveQuadSector::tveQuadSector(tveVertex *aabb_min, tveVertex *aabb_max)
{
    memcpy(&this->aabb_min, aabb_min, sizeof(tveVertex));
    memcpy(&this->aabb_max, aabb_max, sizeof(tveVertex));
    aabb_center.x = (aabb_min->x+aabb_max->x) / 2.0f;
    aabb_center.y = (aabb_min->y+aabb_max->y) / 2.0f;
    aabb_center.z = (aabb_min->z+aabb_max->z) / 2.0f;
}


/*
 *
 */
tveQuadSector::~tveQuadSector()
{
}


//=====================================================================
// tveQuadNode
//=====================================================================

/*
 *
 */
tveQuadNode::tveQuadNode(tveQuadTree *top, tveVertex *aabb_min, tveVertex *aabb_max) : tveQuadSector(aabb_min, aabb_max)
{
    this->top = top;
}


/*
 *
 */
tveQuadNode::~tveQuadNode()
{
    delete upper_left;
    delete upper_right;
    delete lower_left;
    delete lower_right;
}


/*
 *
 */
void tveQuadNode::preCalculate()
{
    visible = ( top->cubeInFrustum(this)!=0 /*&& (aabb_max.y*SCALE_Y>tveVoyager::cv_water->value.floating)*/ );
    upper_left->preCalculate();
    upper_right->preCalculate();
    lower_left->preCalculate();
    lower_right->preCalculate();
}


/*
 *
 */
void tveQuadNode::renderTerrain()
{
    if (!visible)
    {
        return;
    }
    upper_left->renderTerrain();
    upper_right->renderTerrain();
    lower_left->renderTerrain();
    lower_right->renderTerrain();
}


/*
 *
 */
void tveQuadNode::renderGeoMap()
{
    if (!visible)
    {
        return;
    }
    upper_left->renderGeoMap();
    upper_right->renderGeoMap();
    lower_left->renderGeoMap();
    lower_right->renderGeoMap();
}


/*
 *
 */
void tveQuadNode::renderWater()
{
    if (!visible)
    {
        return;
    }
    upper_left->renderWater();
    upper_right->renderWater();
    lower_left->renderWater();
    lower_right->renderWater();
}


//=====================================================================
// tveQuadLeaf
//=====================================================================

/*
 *
 */
tveQuadLeaf::tveQuadLeaf(tveQuadTree *top, tveVertex *aabb_min, tveVertex *aabb_max) : tveQuadSector(aabb_min, aabb_max)
{
    detail2coarse = NULL;
    coarse_color = NULL;
    this->top = top;
    top->addLeaf(this);
    // compute the y value of the bbox
    this->aabb_min.y =  32000;
    this->aabb_max.y = tveVoyager::cv_water->value.floating;//-32000;
    array = top->getVertexArray(aabb_min, aabb_max);
    for (int j=0; j<=LEAF_GRANULARITY; j++)
    {
        for (int i=0; i<=LEAF_GRANULARITY; i++)
        {
            if (array->vertices[j*(LEAF_GRANULARITY+1)+i].y < this->aabb_min.y)
            {
                this->aabb_min.y = array->vertices[j*(LEAF_GRANULARITY+1)+i].y;
            }
            if (array->vertices[j*(LEAF_GRANULARITY+1)+i].y > this->aabb_max.y)
            {
                this->aabb_max.y = array->vertices[j*(LEAF_GRANULARITY+1)+i].y;
            }
        }
    }
    this->aabb_min.y /= SCALE_Y;
    this->aabb_max.y /= SCALE_Y;
}


/*
 *
 */
tveQuadLeaf::~tveQuadLeaf()
{
    if (detail2coarse)
    {
        delete []detail2coarse;
    }
    if (coarse_color)
    {
        delete []coarse_color;
    }
}


/*
 *
 */
void tveQuadLeaf::preCalculate()
{
    visible = ( top->cubeInFrustum(this)!=0 /*&& (aabb_max.y*SCALE_Y>tveVoyager::cv_water->value.floating)*/ );
    if (!visible) return;

    // do geomorphisme
    tveVertex p1 = {tveVoyager::cv_posx->value.floating, tveVoyager::cv_posy->value.floating, -tveVoyager::cv_posz->value.floating};
    tveVertex p2 = {aabb_center.x*SCALE_X, aabb_center.y*SCALE_Y, aabb_center.z*SCALE_Z};
    float dist  = top->distp2p(&p1, &p2);
    float delta = dist - tveVoyager::cv_detail->value.floating;
    //FIXME: if previous frame was geomorphed and the current one isn't, be sure we have the right vertex array
    //       and manage the detail2coarse cache
    array->detail = delta < LEAF_GRANULARITY_SQR*16;
    float factor = MIN(delta/(LEAF_GRANULARITY_SQR*16), 1.0f);
    if ( (factor > 0) && (factor < 1) )
    {
        int offset_x = (int)aabb_min.x;
        int offset_z = (int)aabb_min.z;
        if (!detail2coarse)
        {
            //console->printf("creating detail2coarse");
            detail2coarse = new Sint16[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
            coarse_color  = new Uint8[(LEAF_GRANULARITY+1)*(LEAF_GRANULARITY+1)];
            for (int j=0; j<=LEAF_GRANULARITY; j++)
            {
                for (int i=0; i<=LEAF_GRANULARITY; i++)
                {
                    if (i%2) //  i impair
                    {
                        if (j%2) //  j impair
                        {
                            Sint16 p1 = array->height_map->coarse[(offset_x+i+1)/2][(offset_z+j)/2];
                            Sint16 p2 = array->height_map->coarse[(offset_x+i)/2][(offset_z+j+1)/2];
                            detail2coarse[j*(LEAF_GRANULARITY+1)+i] = (p1+p2)/2;
                        }
                        else
                        {
                            Sint16 p1 = array->height_map->coarse[(offset_x+i)/2][(offset_z+j)/2];
                            Sint16 p2 = array->height_map->coarse[(offset_x+i+1)/2][(offset_z+j)/2];
                            detail2coarse[j*(LEAF_GRANULARITY+1)+i] = (p1+p2)/2;
                        }
                    }
                    else //  i pair
                    {
                        if (j%2) //  j pair
                        {
                            Sint16 p1 = array->height_map->coarse[(offset_x+i)/2][(offset_z+j)/2];
                            Sint16 p2 = array->height_map->coarse[(offset_x+i)/2][(offset_z+j+1)/2];
                            detail2coarse[j*(LEAF_GRANULARITY+1)+i] = (p1+p2)/2;
                        }
                        else
                        {
                            Sint16 p1 = array->height_map->coarse[(offset_x+i+1)/2][(offset_z+j)/2];
                            Sint16 p2 = array->height_map->coarse[(offset_x+i+1)/2][(offset_z+j+1)/2];
                            detail2coarse[j*(LEAF_GRANULARITY+1)+i] = (p1+p2)/2;
                        }
                    }
                }
            }
        }
        for (int z=1; z<LEAF_GRANULARITY; z++)
        {
            for (int x=1; x<LEAF_GRANULARITY; x++)
            {
                array->vertices[z*(LEAF_GRANULARITY+1)+x].y = (1-factor) * array->height_map->detail[offset_x+x][offset_z+z] * SCALE_Y;
                array->vertices[z*(LEAF_GRANULARITY+1)+x].y += (factor) * detail2coarse[z*(LEAF_GRANULARITY+1)+x] * SCALE_Y;
            }
        }
    }
    else
    {
        if (detail2coarse)
        {
            //console->printf("deleting detail2coarse");
            //delete []detail2coarse;
            //detail2coarse = NULL;
        }
    }
}


/*
 *
 */
void tveQuadLeaf::renderTerrain()
{
    if (visible)
    {
        tvePattern *pattern = top->getVertexArray(&aabb_min, &aabb_max);

        //FIXME: put that code here or in pattern->render() ?
        if (tveVoyager::cv_posy->value.floating < tveVoyager::cv_water->value.floating)
        {
            float col = MAX(1-(tveVoyager::cv_water->value.floating-tveVoyager::cv_posy->value.floating)*0.4f, 0.5f);
            glColor4f(col, col, 1.0f, 1.0f);
        }
        else
        {
            glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        }
        pattern->render(aabb_max.y, top->getLOD(this));
        top->num_sect++;
        top->num_poly += (array->detail ? 2*LEAF_GRANULARITY*LEAF_GRANULARITY*2 : 2*LEAF_GRANULARITY*LEAF_GRANULARITY/2) + ((aabb_max.y > 128.0f) ? LEAF_GRANULARITY*LEAF_GRANULARITY*2 : 0);
    }
}


/*
 *
 */
void tveQuadLeaf::renderGeoMap()
{
    if (!visible)
    {
        return;
    }

    tveHeightMap *height_map = array->height_map;

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, array->height_map->tex[TREE]);

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    for (int j=(int)aabb_min.z; j<(int)aabb_max.z; j++)
    {
        for (int i=(int)aabb_min.x; i<(int)aabb_max.x; i++)
        {
            if (height_map->geomap[i][j] & GEO_TREE)
            {
                float th = height_map->detail[i][j]*SCALE_Y;
                glBegin(GL_TRIANGLE_STRIP);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
                  glVertex3f(0.0f+i*2, th, 1.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
                  glVertex3f(0.0f+i*2, 3.0f+th, 1.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
                  glVertex3f(2.0f+i*2, th, 1.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
                  glVertex3f(2.0f+i*2, 3.0f+th, 1.0f+j*2);
                glEnd();
                glBegin(GL_TRIANGLE_STRIP);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
                  glVertex3f(1.0f+i*2, th, 0.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
                  glVertex3f(1.0f+i*2, 3.0f+th, 0.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
                  glVertex3f(1.0f+i*2, th, 2.0f+j*2);
                  glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
                  glVertex3f(1.0f+i*2, 3.0f+th, 2.0f+j*2);
                glEnd();
            }
        }
    }
}


/*
 *
 */
void tveQuadLeaf::renderWater()
{
    float wh = tveVoyager::cv_water->value.floating;

    if (!visible || (aabb_min.y*SCALE_Y > wh))
    {
        return;
    }
    tveHeightMap *height_map = array->height_map;
    float wind = top->wind;
    // GL_TEXTURE0_ARB = water, GL_TEXTURE1_ARB = sky
    /*
    glBegin(GL_TRIANGLE_STRIP);
        glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*aabb_min.x+wind*10.0f, 0.5f*aabb_min.z+wind*10.0f);
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*aabb_min.x/height_map->width+wind, 2.0f*aabb_min.z/height_map->height+wind);
        glVertex3f(aabb_min.x*SCALE_X, wh, aabb_min.z*SCALE_Z);
        glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*aabb_min.x+wind*10.0f, 0.5f*aabb_max.z+wind*10.0f);
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*aabb_min.x/height_map->width+wind, 2.0f*aabb_max.z/height_map->height+wind);
        glVertex3f(aabb_min.x*SCALE_X, wh, aabb_max.z*SCALE_Z);
        glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*aabb_max.x+wind*10.0f, 0.5f*aabb_min.z+wind*10.0f);
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*aabb_max.x/height_map->width+wind, 2.0f*aabb_min.z/height_map->height+wind);
        glVertex3f(aabb_max.x*SCALE_X, wh, aabb_min.z*SCALE_Z);
        glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*aabb_max.x+wind*10.0f, 0.5f*aabb_max.z+wind*10.0f);
        glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*aabb_max.x/height_map->width+wind, 2.0f*aabb_max.z/height_map->height+wind);
        glVertex3f(aabb_max.x*SCALE_X, wh, aabb_max.z*SCALE_Z);
    glEnd();
    */
    for (int i=0; i<2; i++)
    {
        for (int j=0; j<2; j++)
        {
            float minx = aabb_min.x+(i*(aabb_max.x-aabb_min.x))/2;
            float maxx = minx + (aabb_max.x-aabb_min.x)/2;
            float minz = aabb_min.z+(j*(aabb_max.z-aabb_min.z))/2;
            float maxz = minz + (aabb_max.z-aabb_min.z)/2;
            glBegin(GL_TRIANGLE_STRIP);
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*minx+wind*10.0f, 0.5f*minz+wind*10.0f);
            glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*minx/height_map->width+wind, 2.0f*minz/height_map->height+wind);
            glVertex3f(minx*SCALE_X, wh, minz*SCALE_Z);
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*minx+wind*10.0f, 0.5f*maxz+wind*10.0f);
            glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*minx/height_map->width+wind, 2.0f*maxz/height_map->height+wind);
            glVertex3f(minx*SCALE_X, wh, maxz*SCALE_Z);
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*maxx+wind*10.0f, 0.5f*minz+wind*10.0f);
            glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*maxx/height_map->width+wind, 2.0f*minz/height_map->height+wind);
            glVertex3f(maxx*SCALE_X, wh, minz*SCALE_Z);
            glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.5f*maxx+wind*10.0f, 0.5f*maxz+wind*10.0f);
            glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f*maxx/height_map->width+wind, 2.0f*maxz/height_map->height+wind);
            glVertex3f(maxx*SCALE_X, wh, maxz*SCALE_Z);
            glEnd();
        }
    }
}



/*
 *
 */
bool tveQuadLeaf::getDetail()
{
    return array->detail;
}
